from __future__ import annotations

from abc import ABC, abstractmethod


class MetadataHookInterface(ABC):  # no cov
    """
    Example usage:

    === ":octicons-file-code-16: plugin.py"

        ```python
        from hatchling.metadata.plugin.interface import MetadataHookInterface


        class SpecialMetadataHook(MetadataHookInterface):
            PLUGIN_NAME = 'special'
            ...
        ```

    === ":octicons-file-code-16: hooks.py"

        ```python
        from hatchling.plugin import hookimpl

        from .plugin import SpecialMetadataHook


        @hookimpl
        def hatch_register_metadata_hook():
            return SpecialMetadataHook
        ```
    """

    PLUGIN_NAME = ''
    """The name used for selection."""

    def __init__(self, root, config):
        self.__root = root
        self.__config = config

    @property
    def root(self):
        """
        The root of the project tree.
        """
        return self.__root

    @property
    def config(self):
        """
        The hook configuration.

        === ":octicons-file-code-16: pyproject.toml"

            ```toml
            [tool.hatch.metadata.hooks.<PLUGIN_NAME>]
            ```

        === ":octicons-file-code-16: hatch.toml"

            ```toml
            [metadata.hooks.<PLUGIN_NAME>]
            ```
        """
        return self.__config

    @abstractmethod
    def update(self, metadata: dict):
        """
        This updates the metadata mapping of the `project` table in-place.
        """

    def get_known_classifiers(self) -> list[str]:
        """
        This returns extra classifiers that should be considered valid in addition to the ones known to PyPI.
        """
        return []



import pluggy


class PluginManager:
    def __init__(self):
        self.manager = pluggy.PluginManager('hatch')
        self.third_party_plugins = ThirdPartyPlugins(self.manager)
        self.initialized = False

    def initialize(self):
        from hatchling.plugin import specs

        self.manager.add_hookspecs(specs)

    def __getattr__(self, name):
        if not self.initialized:
            self.initialize()
            self.initialized = True

        hook_name = f'hatch_register_{name}'
        getattr(self, hook_name, None)()

        register = ClassRegister(getattr(self.manager.hook, hook_name), 'PLUGIN_NAME', self.third_party_plugins)
        setattr(self, name, register)
        return register

    def hatch_register_version_source(self):
        from hatchling.version.source.plugin import hooks

        self.manager.register(hooks)

    def hatch_register_version_scheme(self):
        from hatchling.version.scheme.plugin import hooks

        self.manager.register(hooks)

    def hatch_register_builder(self):
        from hatchling.builders.plugin import hooks

        self.manager.register(hooks)

    def hatch_register_build_hook(self):
        from hatchling.builders.hooks.plugin import hooks

        self.manager.register(hooks)

    def hatch_register_metadata_hook(self):
        from hatchling.metadata.plugin import hooks

        self.manager.register(hooks)


class ClassRegister:
    def __init__(self, registration_method, identifier, third_party_plugins):
        self.registration_method = registration_method
        self.identifier = identifier
        self.third_party_plugins = third_party_plugins

    def collect(self, include_third_party=True):
        if include_third_party and not self.third_party_plugins.loaded:
            self.third_party_plugins.load()

        classes = {}

        for registered_classes in self.registration_method():
            if not isinstance(registered_classes, list):
                registered_classes = [registered_classes]

            for registered_class in registered_classes:
                name = getattr(registered_class, self.identifier, None)
                if not name:  # no cov
                    raise ValueError(f'Class `{registered_class.__name__}` does not have a {name} attribute.')
                elif name in classes:  # no cov
                    raise ValueError(
                        f'Class `{registered_class.__name__}` defines its name as `{name}` but '
                        f'that name is already used by `{classes[name].__name__}`.'
                    )

                classes[name] = registered_class

        return classes

    def get(self, name):
        if not self.third_party_plugins.loaded:
            classes = self.collect(include_third_party=False)
            if name in classes:
                return classes[name]

        return self.collect().get(name)


class ThirdPartyPlugins:
    def __init__(self, manager):
        self.manager = manager
        self.loaded = False

    def load(self):
        self.manager.load_setuptools_entrypoints('hatch')
        self.loaded = True



import argparse

from hatchling.cli.build import build_command
from hatchling.cli.dep import dep_command
from hatchling.cli.version import version_command


def hatchling():
    parser = argparse.ArgumentParser(prog='hatchling', allow_abbrev=False)
    subparsers = parser.add_subparsers(required=True)

    defaults = {'metavar': ''}

    build_command(subparsers, defaults)
    dep_command(subparsers, defaults)
    version_command(subparsers, defaults)

    kwargs = vars(parser.parse_args())
    command = kwargs.pop('func')
    command(**kwargs)



import os
import shutil
from base64 import urlsafe_b64encode


def replace_file(src, dst):
    try:
        os.replace(src, dst)
    # Happens when on different filesystems like /tmp or caused by layering in containers
    except OSError:
        shutil.copy2(src, dst)
        os.remove(src)


def safe_walk(path):
    seen = set()
    for root, dirs, files in os.walk(path, followlinks=True):
        stat = os.stat(root)
        identifier = stat.st_dev, stat.st_ino
        if identifier in seen:
            del dirs[:]
            continue

        seen.add(identifier)
        yield root, dirs, files


def get_known_python_major_versions():
    return map(str, sorted((2, 3)))


def get_relative_path(path, start):
    relative_path = os.path.relpath(path, start)

    # First iteration of `os.walk`
    if relative_path == '.':
        return ''

    return relative_path


def normalize_relative_path(path):
    return os.path.normpath(path).strip(os.sep)


def normalize_relative_directory(path):
    return normalize_relative_path(path) + os.sep


def normalize_inclusion_map(inclusion_map, root):
    normalized_inclusion_map = {}

    for source, relative_path in inclusion_map.items():
        source = os.path.expanduser(os.path.normpath(source))
        if not os.path.isabs(source):
            source = os.path.abspath(os.path.join(root, source))

        normalized_inclusion_map[source] = normalize_relative_path(relative_path)

    return dict(sorted(normalized_inclusion_map.items(), key=lambda item: (item[1].count(os.sep), item[1], item[0])))


def normalize_archive_path(path):
    if os.sep != '/':
        return path.replace(os.sep, '/')

    return path


def format_file_hash(digest):
    # https://peps.python.org/pep-0427/#signed-wheel-files
    return urlsafe_b64encode(digest).decode('ascii').rstrip('=')


def get_reproducible_timestamp():
    """
    Returns an `int` derived from the `SOURCE_DATE_EPOCH` environment variable; see
    https://reproducible-builds.org/specs/source-date-epoch/.

    The default value will always be: `1580601600`
    """
    return int(os.environ.get('SOURCE_DATE_EPOCH', '1580601600'))


def normalize_file_permissions(st_mode):
    """
    https://github.com/takluyver/flit/blob/6a2a8c6462e49f584941c667b70a6f48a7b3f9ab/flit_core/flit_core/common.py#L257

    Normalize the permission bits in the st_mode field from stat to 644/755.

    Popular VCSs only track whether a file is executable or not. The exact
    permissions can vary on systems with different umasks. Normalizing
    to 644 (non executable) or 755 (executable) makes builds more reproducible.
    """
    # Set 644 permissions, leaving higher bits of st_mode unchanged
    new_mode = (st_mode | 0o644) & ~0o133
    if st_mode & 0o100:  # no cov
        new_mode |= 0o111  # Executable: 644 -> 755
    return new_mode


def set_zip_info_mode(zip_info, mode=0o644):
    """
    https://github.com/takluyver/flit/commit/3889583719888aef9f28baaa010e698cb7884904
    """
    zip_info.external_attr = mode << 16



VERSION = '2022.8.7'

SORTED_CLASSIFIERS = [
    'Development Status :: 1 - Planning',
    'Development Status :: 2 - Pre-Alpha',
    'Development Status :: 3 - Alpha',
    'Development Status :: 4 - Beta',
    'Development Status :: 5 - Production/Stable',
    'Development Status :: 6 - Mature',
    'Development Status :: 7 - Inactive',
    'Environment :: Console',
    'Environment :: Console :: Curses',
    'Environment :: Console :: Framebuffer',
    'Environment :: Console :: Newt',
    'Environment :: Console :: svgalib',
    'Environment :: GPU',
    'Environment :: GPU :: NVIDIA CUDA',
    'Environment :: GPU :: NVIDIA CUDA :: 1.0',
    'Environment :: GPU :: NVIDIA CUDA :: 1.1',
    'Environment :: GPU :: NVIDIA CUDA :: 2.0',
    'Environment :: GPU :: NVIDIA CUDA :: 2.1',
    'Environment :: GPU :: NVIDIA CUDA :: 2.2',
    'Environment :: GPU :: NVIDIA CUDA :: 2.3',
    'Environment :: GPU :: NVIDIA CUDA :: 3.0',
    'Environment :: GPU :: NVIDIA CUDA :: 3.1',
    'Environment :: GPU :: NVIDIA CUDA :: 3.2',
    'Environment :: GPU :: NVIDIA CUDA :: 4.0',
    'Environment :: GPU :: NVIDIA CUDA :: 4.1',
    'Environment :: GPU :: NVIDIA CUDA :: 4.2',
    'Environment :: GPU :: NVIDIA CUDA :: 5.0',
    'Environment :: GPU :: NVIDIA CUDA :: 5.5',
    'Environment :: GPU :: NVIDIA CUDA :: 6.0',
    'Environment :: GPU :: NVIDIA CUDA :: 6.5',
    'Environment :: GPU :: NVIDIA CUDA :: 7.0',
    'Environment :: GPU :: NVIDIA CUDA :: 7.5',
    'Environment :: GPU :: NVIDIA CUDA :: 8.0',
    'Environment :: GPU :: NVIDIA CUDA :: 9.0',
    'Environment :: GPU :: NVIDIA CUDA :: 9.1',
    'Environment :: GPU :: NVIDIA CUDA :: 9.2',
    'Environment :: GPU :: NVIDIA CUDA :: 10.0',
    'Environment :: GPU :: NVIDIA CUDA :: 10.1',
    'Environment :: GPU :: NVIDIA CUDA :: 10.2',
    'Environment :: GPU :: NVIDIA CUDA :: 11.0',
    'Environment :: GPU :: NVIDIA CUDA :: 11.1',
    'Environment :: GPU :: NVIDIA CUDA :: 11.2',
    'Environment :: GPU :: NVIDIA CUDA :: 11.3',
    'Environment :: GPU :: NVIDIA CUDA :: 11.4',
    'Environment :: GPU :: NVIDIA CUDA :: 11.5',
    'Environment :: GPU :: NVIDIA CUDA :: 11.6',
    'Environment :: GPU :: NVIDIA CUDA :: 11.7',
    "Environment :: Handhelds/PDA's",
    'Environment :: MacOS X',
    'Environment :: MacOS X :: Aqua',
    'Environment :: MacOS X :: Carbon',
    'Environment :: MacOS X :: Cocoa',
    'Environment :: No Input/Output (Daemon)',
    'Environment :: OpenStack',
    'Environment :: Other Environment',
    'Environment :: Plugins',
    'Environment :: Web Environment',
    'Environment :: Web Environment :: Buffet',
    'Environment :: Web Environment :: Mozilla',
    'Environment :: Web Environment :: ToscaWidgets',
    'Environment :: Win32 (MS Windows)',
    'Environment :: X11 Applications',
    'Environment :: X11 Applications :: GTK',
    'Environment :: X11 Applications :: Gnome',
    'Environment :: X11 Applications :: KDE',
    'Environment :: X11 Applications :: Qt',
    'Framework :: AWS CDK',
    'Framework :: AWS CDK :: 1',
    'Framework :: AWS CDK :: 2',
    'Framework :: AiiDA',
    'Framework :: Ansible',
    'Framework :: AnyIO',
    'Framework :: Apache Airflow',
    'Framework :: Apache Airflow :: Provider',
    'Framework :: AsyncIO',
    'Framework :: BEAT',
    'Framework :: BFG',
    'Framework :: Bob',
    'Framework :: Bottle',
    'Framework :: Buildout',
    'Framework :: Buildout :: Extension',
    'Framework :: Buildout :: Recipe',
    'Framework :: CastleCMS',
    'Framework :: CastleCMS :: Theme',
    'Framework :: Celery',
    'Framework :: Chandler',
    'Framework :: CherryPy',
    'Framework :: CubicWeb',
    'Framework :: Dash',
    'Framework :: Datasette',
    'Framework :: Django',
    'Framework :: Django :: 1',
    'Framework :: Django :: 1.4',
    'Framework :: Django :: 1.5',
    'Framework :: Django :: 1.6',
    'Framework :: Django :: 1.7',
    'Framework :: Django :: 1.8',
    'Framework :: Django :: 1.9',
    'Framework :: Django :: 1.10',
    'Framework :: Django :: 1.11',
    'Framework :: Django :: 2',
    'Framework :: Django :: 2.0',
    'Framework :: Django :: 2.1',
    'Framework :: Django :: 2.2',
    'Framework :: Django :: 3',
    'Framework :: Django :: 3.0',
    'Framework :: Django :: 3.1',
    'Framework :: Django :: 3.2',
    'Framework :: Django :: 4',
    'Framework :: Django :: 4.0',
    'Framework :: Django :: 4.1',
    'Framework :: Django CMS',
    'Framework :: Django CMS :: 3.4',
    'Framework :: Django CMS :: 3.5',
    'Framework :: Django CMS :: 3.6',
    'Framework :: Django CMS :: 3.7',
    'Framework :: Django CMS :: 3.8',
    'Framework :: Django CMS :: 3.9',
    'Framework :: Django CMS :: 3.10',
    'Framework :: FastAPI',
    'Framework :: Flake8',
    'Framework :: Flask',
    'Framework :: Hatch',
    'Framework :: Hypothesis',
    'Framework :: IDLE',
    'Framework :: IPython',
    'Framework :: Jupyter',
    'Framework :: Jupyter :: JupyterLab',
    'Framework :: Jupyter :: JupyterLab :: 1',
    'Framework :: Jupyter :: JupyterLab :: 2',
    'Framework :: Jupyter :: JupyterLab :: 3',
    'Framework :: Jupyter :: JupyterLab :: 4',
    'Framework :: Jupyter :: JupyterLab :: Extensions',
    'Framework :: Jupyter :: JupyterLab :: Extensions :: Mime Renderers',
    'Framework :: Jupyter :: JupyterLab :: Extensions :: Prebuilt',
    'Framework :: Jupyter :: JupyterLab :: Extensions :: Themes',
    'Framework :: Kedro',
    'Framework :: Lektor',
    'Framework :: Masonite',
    'Framework :: Matplotlib',
    'Framework :: Nengo',
    'Framework :: Odoo',
    'Framework :: Odoo :: 8.0',
    'Framework :: Odoo :: 9.0',
    'Framework :: Odoo :: 10.0',
    'Framework :: Odoo :: 11.0',
    'Framework :: Odoo :: 12.0',
    'Framework :: Odoo :: 13.0',
    'Framework :: Odoo :: 14.0',
    'Framework :: Odoo :: 15.0',
    'Framework :: Opps',
    'Framework :: Paste',
    'Framework :: Pelican',
    'Framework :: Pelican :: Plugins',
    'Framework :: Pelican :: Themes',
    'Framework :: Plone',
    'Framework :: Plone :: 3.2',
    'Framework :: Plone :: 3.3',
    'Framework :: Plone :: 4.0',
    'Framework :: Plone :: 4.1',
    'Framework :: Plone :: 4.2',
    'Framework :: Plone :: 4.3',
    'Framework :: Plone :: 5.0',
    'Framework :: Plone :: 5.1',
    'Framework :: Plone :: 5.2',
    'Framework :: Plone :: 5.3',
    'Framework :: Plone :: 6.0',
    'Framework :: Plone :: Addon',
    'Framework :: Plone :: Core',
    'Framework :: Plone :: Theme',
    'Framework :: Pylons',
    'Framework :: Pyramid',
    'Framework :: Pytest',
    'Framework :: Review Board',
    'Framework :: Robot Framework',
    'Framework :: Robot Framework :: Library',
    'Framework :: Robot Framework :: Tool',
    'Framework :: Scrapy',
    'Framework :: Setuptools Plugin',
    'Framework :: Sphinx',
    'Framework :: Sphinx :: Extension',
    'Framework :: Sphinx :: Theme',
    'Framework :: Trac',
    'Framework :: Trio',
    'Framework :: Tryton',
    'Framework :: TurboGears',
    'Framework :: TurboGears :: Applications',
    'Framework :: TurboGears :: Widgets',
    'Framework :: Twisted',
    'Framework :: Wagtail',
    'Framework :: Wagtail :: 1',
    'Framework :: Wagtail :: 2',
    'Framework :: Wagtail :: 3',
    'Framework :: ZODB',
    'Framework :: Zope',
    'Framework :: Zope2',
    'Framework :: Zope3',
    'Framework :: Zope :: 2',
    'Framework :: Zope :: 3',
    'Framework :: Zope :: 4',
    'Framework :: Zope :: 5',
    'Framework :: aiohttp',
    'Framework :: cocotb',
    'Framework :: napari',
    'Framework :: tox',
    'Intended Audience :: Customer Service',
    'Intended Audience :: Developers',
    'Intended Audience :: Education',
    'Intended Audience :: End Users/Desktop',
    'Intended Audience :: Financial and Insurance Industry',
    'Intended Audience :: Healthcare Industry',
    'Intended Audience :: Information Technology',
    'Intended Audience :: Legal Industry',
    'Intended Audience :: Manufacturing',
    'Intended Audience :: Other Audience',
    'Intended Audience :: Religion',
    'Intended Audience :: Science/Research',
    'Intended Audience :: System Administrators',
    'Intended Audience :: Telecommunications Industry',
    'License :: Aladdin Free Public License (AFPL)',
    'License :: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication',
    'License :: CeCILL-B Free Software License Agreement (CECILL-B)',
    'License :: CeCILL-C Free Software License Agreement (CECILL-C)',
    'License :: DFSG approved',
    'License :: Eiffel Forum License (EFL)',
    'License :: Free For Educational Use',
    'License :: Free For Home Use',
    'License :: Free To Use But Restricted',
    'License :: Free for non-commercial use',
    'License :: Freely Distributable',
    'License :: Freeware',
    'License :: GUST Font License 1.0',
    'License :: GUST Font License 2006-09-30',
    'License :: Netscape Public License (NPL)',
    'License :: Nokia Open Source License (NOKOS)',
    'License :: OSI Approved',
    'License :: OSI Approved :: Academic Free License (AFL)',
    'License :: OSI Approved :: Apache Software License',
    'License :: OSI Approved :: Apple Public Source License',
    'License :: OSI Approved :: Artistic License',
    'License :: OSI Approved :: Attribution Assurance License',
    'License :: OSI Approved :: BSD License',
    'License :: OSI Approved :: Boost Software License 1.0 (BSL-1.0)',
    'License :: OSI Approved :: CEA CNRS Inria Logiciel Libre License, version 2.1 (CeCILL-2.1)',
    'License :: OSI Approved :: Common Development and Distribution License 1.0 (CDDL-1.0)',
    'License :: OSI Approved :: Common Public License',
    'License :: OSI Approved :: Eclipse Public License 1.0 (EPL-1.0)',
    'License :: OSI Approved :: Eclipse Public License 2.0 (EPL-2.0)',
    'License :: OSI Approved :: Eiffel Forum License',
    'License :: OSI Approved :: European Union Public Licence 1.0 (EUPL 1.0)',
    'License :: OSI Approved :: European Union Public Licence 1.1 (EUPL 1.1)',
    'License :: OSI Approved :: European Union Public Licence 1.2 (EUPL 1.2)',
    'License :: OSI Approved :: GNU Affero General Public License v3',
    'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
    'License :: OSI Approved :: GNU Free Documentation License (FDL)',
    'License :: OSI Approved :: GNU General Public License (GPL)',
    'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
    'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
    'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
    'License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)',
    'License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)',
    'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
    'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
    'License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)',
    'License :: OSI Approved :: Historical Permission Notice and Disclaimer (HPND)',
    'License :: OSI Approved :: IBM Public License',
    'License :: OSI Approved :: ISC License (ISCL)',
    'License :: OSI Approved :: Intel Open Source License',
    'License :: OSI Approved :: Jabber Open Source License',
    'License :: OSI Approved :: MIT License',
    'License :: OSI Approved :: MIT No Attribution License (MIT-0)',
    'License :: OSI Approved :: MITRE Collaborative Virtual Workspace License (CVW)',
    'License :: OSI Approved :: MirOS License (MirOS)',
    'License :: OSI Approved :: Motosoto License',
    'License :: OSI Approved :: Mozilla Public License 1.0 (MPL)',
    'License :: OSI Approved :: Mozilla Public License 1.1 (MPL 1.1)',
    'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
    'License :: OSI Approved :: Mulan Permissive Software License v2 (MulanPSL-2.0)',
    'License :: OSI Approved :: Nethack General Public License',
    'License :: OSI Approved :: Nokia Open Source License',
    'License :: OSI Approved :: Open Group Test Suite License',
    'License :: OSI Approved :: Open Software License 3.0 (OSL-3.0)',
    'License :: OSI Approved :: PostgreSQL License',
    'License :: OSI Approved :: Python License (CNRI Python License)',
    'License :: OSI Approved :: Python Software Foundation License',
    'License :: OSI Approved :: Qt Public License (QPL)',
    'License :: OSI Approved :: Ricoh Source Code Public License',
    'License :: OSI Approved :: SIL Open Font License 1.1 (OFL-1.1)',
    'License :: OSI Approved :: Sleepycat License',
    'License :: OSI Approved :: Sun Industry Standards Source License (SISSL)',
    'License :: OSI Approved :: Sun Public License',
    'License :: OSI Approved :: The Unlicense (Unlicense)',
    'License :: OSI Approved :: Universal Permissive License (UPL)',
    'License :: OSI Approved :: University of Illinois/NCSA Open Source License',
    'License :: OSI Approved :: Vovida Software License 1.0',
    'License :: OSI Approved :: W3C License',
    'License :: OSI Approved :: X.Net License',
    'License :: OSI Approved :: Zope Public License',
    'License :: OSI Approved :: zlib/libpng License',
    'License :: Other/Proprietary License',
    'License :: Public Domain',
    'License :: Repoze Public License',
    'Natural Language :: Afrikaans',
    'Natural Language :: Arabic',
    'Natural Language :: Basque',
    'Natural Language :: Bengali',
    'Natural Language :: Bosnian',
    'Natural Language :: Bulgarian',
    'Natural Language :: Cantonese',
    'Natural Language :: Catalan',
    'Natural Language :: Chinese (Simplified)',
    'Natural Language :: Chinese (Traditional)',
    'Natural Language :: Croatian',
    'Natural Language :: Czech',
    'Natural Language :: Danish',
    'Natural Language :: Dutch',
    'Natural Language :: English',
    'Natural Language :: Esperanto',
    'Natural Language :: Finnish',
    'Natural Language :: French',
    'Natural Language :: Galician',
    'Natural Language :: German',
    'Natural Language :: Greek',
    'Natural Language :: Hebrew',
    'Natural Language :: Hindi',
    'Natural Language :: Hungarian',
    'Natural Language :: Icelandic',
    'Natural Language :: Indonesian',
    'Natural Language :: Irish',
    'Natural Language :: Italian',
    'Natural Language :: Japanese',
    'Natural Language :: Javanese',
    'Natural Language :: Korean',
    'Natural Language :: Latin',
    'Natural Language :: Latvian',
    'Natural Language :: Lithuanian',
    'Natural Language :: Macedonian',
    'Natural Language :: Malay',
    'Natural Language :: Marathi',
    'Natural Language :: Nepali',
    'Natural Language :: Norwegian',
    'Natural Language :: Panjabi',
    'Natural Language :: Persian',
    'Natural Language :: Polish',
    'Natural Language :: Portuguese',
    'Natural Language :: Portuguese (Brazilian)',
    'Natural Language :: Romanian',
    'Natural Language :: Russian',
    'Natural Language :: Serbian',
    'Natural Language :: Slovak',
    'Natural Language :: Slovenian',
    'Natural Language :: Spanish',
    'Natural Language :: Swedish',
    'Natural Language :: Tamil',
    'Natural Language :: Telugu',
    'Natural Language :: Thai',
    'Natural Language :: Tibetan',
    'Natural Language :: Turkish',
    'Natural Language :: Ukrainian',
    'Natural Language :: Urdu',
    'Natural Language :: Vietnamese',
    'Operating System :: Android',
    'Operating System :: BeOS',
    'Operating System :: MacOS',
    'Operating System :: MacOS :: MacOS 9',
    'Operating System :: MacOS :: MacOS X',
    'Operating System :: Microsoft',
    'Operating System :: Microsoft :: MS-DOS',
    'Operating System :: Microsoft :: Windows',
    'Operating System :: Microsoft :: Windows :: Windows 3.1 or Earlier',
    'Operating System :: Microsoft :: Windows :: Windows 7',
    'Operating System :: Microsoft :: Windows :: Windows 8',
    'Operating System :: Microsoft :: Windows :: Windows 8.1',
    'Operating System :: Microsoft :: Windows :: Windows 10',
    'Operating System :: Microsoft :: Windows :: Windows 11',
    'Operating System :: Microsoft :: Windows :: Windows 95/98/2000',
    'Operating System :: Microsoft :: Windows :: Windows CE',
    'Operating System :: Microsoft :: Windows :: Windows NT/2000',
    'Operating System :: Microsoft :: Windows :: Windows Server 2003',
    'Operating System :: Microsoft :: Windows :: Windows Server 2008',
    'Operating System :: Microsoft :: Windows :: Windows Vista',
    'Operating System :: Microsoft :: Windows :: Windows XP',
    'Operating System :: OS Independent',
    'Operating System :: OS/2',
    'Operating System :: Other OS',
    'Operating System :: PDA Systems',
    'Operating System :: POSIX',
    'Operating System :: POSIX :: AIX',
    'Operating System :: POSIX :: BSD',
    'Operating System :: POSIX :: BSD :: BSD/OS',
    'Operating System :: POSIX :: BSD :: FreeBSD',
    'Operating System :: POSIX :: BSD :: NetBSD',
    'Operating System :: POSIX :: BSD :: OpenBSD',
    'Operating System :: POSIX :: GNU Hurd',
    'Operating System :: POSIX :: HP-UX',
    'Operating System :: POSIX :: IRIX',
    'Operating System :: POSIX :: Linux',
    'Operating System :: POSIX :: Other',
    'Operating System :: POSIX :: SCO',
    'Operating System :: POSIX :: SunOS/Solaris',
    'Operating System :: PalmOS',
    'Operating System :: RISC OS',
    'Operating System :: Unix',
    'Operating System :: iOS',
    'Programming Language :: APL',
    'Programming Language :: ASP',
    'Programming Language :: Ada',
    'Programming Language :: Assembly',
    'Programming Language :: Awk',
    'Programming Language :: Basic',
    'Programming Language :: C',
    'Programming Language :: C#',
    'Programming Language :: C++',
    'Programming Language :: Cold Fusion',
    'Programming Language :: Cython',
    'Programming Language :: D',
    'Programming Language :: Delphi/Kylix',
    'Programming Language :: Dylan',
    'Programming Language :: Eiffel',
    'Programming Language :: Emacs-Lisp',
    'Programming Language :: Erlang',
    'Programming Language :: Euler',
    'Programming Language :: Euphoria',
    'Programming Language :: F#',
    'Programming Language :: Forth',
    'Programming Language :: Fortran',
    'Programming Language :: Haskell',
    'Programming Language :: Java',
    'Programming Language :: JavaScript',
    'Programming Language :: Kotlin',
    'Programming Language :: Lisp',
    'Programming Language :: Logo',
    'Programming Language :: ML',
    'Programming Language :: Modula',
    'Programming Language :: OCaml',
    'Programming Language :: Object Pascal',
    'Programming Language :: Objective C',
    'Programming Language :: Other',
    'Programming Language :: Other Scripting Engines',
    'Programming Language :: PHP',
    'Programming Language :: PL/SQL',
    'Programming Language :: PROGRESS',
    'Programming Language :: Pascal',
    'Programming Language :: Perl',
    'Programming Language :: Pike',
    'Programming Language :: Pliant',
    'Programming Language :: Prolog',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2 :: Only',
    'Programming Language :: Python :: 2.3',
    'Programming Language :: Python :: 2.4',
    'Programming Language :: Python :: 2.5',
    'Programming Language :: Python :: 2.6',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3 :: Only',
    'Programming Language :: Python :: 3.0',
    'Programming Language :: Python :: 3.1',
    'Programming Language :: Python :: 3.2',
    'Programming Language :: Python :: 3.3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Programming Language :: Python :: 3.11',
    'Programming Language :: Python :: 3.12',
    'Programming Language :: Python :: Implementation',
    'Programming Language :: Python :: Implementation :: CPython',
    'Programming Language :: Python :: Implementation :: IronPython',
    'Programming Language :: Python :: Implementation :: Jython',
    'Programming Language :: Python :: Implementation :: MicroPython',
    'Programming Language :: Python :: Implementation :: PyPy',
    'Programming Language :: Python :: Implementation :: Stackless',
    'Programming Language :: R',
    'Programming Language :: REBOL',
    'Programming Language :: Rexx',
    'Programming Language :: Ruby',
    'Programming Language :: Rust',
    'Programming Language :: SQL',
    'Programming Language :: Scheme',
    'Programming Language :: Simula',
    'Programming Language :: Smalltalk',
    'Programming Language :: Tcl',
    'Programming Language :: Unix Shell',
    'Programming Language :: Visual Basic',
    'Programming Language :: XBasic',
    'Programming Language :: YACC',
    'Programming Language :: Zope',
    'Topic :: Adaptive Technologies',
    'Topic :: Artistic Software',
    'Topic :: Communications',
    'Topic :: Communications :: BBS',
    'Topic :: Communications :: Chat',
    'Topic :: Communications :: Chat :: ICQ',
    'Topic :: Communications :: Chat :: Internet Relay Chat',
    'Topic :: Communications :: Chat :: Unix Talk',
    'Topic :: Communications :: Conferencing',
    'Topic :: Communications :: Email',
    'Topic :: Communications :: Email :: Address Book',
    'Topic :: Communications :: Email :: Email Clients (MUA)',
    'Topic :: Communications :: Email :: Filters',
    'Topic :: Communications :: Email :: Mail Transport Agents',
    'Topic :: Communications :: Email :: Mailing List Servers',
    'Topic :: Communications :: Email :: Post-Office',
    'Topic :: Communications :: Email :: Post-Office :: IMAP',
    'Topic :: Communications :: Email :: Post-Office :: POP3',
    'Topic :: Communications :: FIDO',
    'Topic :: Communications :: Fax',
    'Topic :: Communications :: File Sharing',
    'Topic :: Communications :: File Sharing :: Gnutella',
    'Topic :: Communications :: File Sharing :: Napster',
    'Topic :: Communications :: Ham Radio',
    'Topic :: Communications :: Internet Phone',
    'Topic :: Communications :: Telephony',
    'Topic :: Communications :: Usenet News',
    'Topic :: Database',
    'Topic :: Database :: Database Engines/Servers',
    'Topic :: Database :: Front-Ends',
    'Topic :: Desktop Environment',
    'Topic :: Desktop Environment :: File Managers',
    'Topic :: Desktop Environment :: GNUstep',
    'Topic :: Desktop Environment :: Gnome',
    'Topic :: Desktop Environment :: K Desktop Environment (KDE)',
    'Topic :: Desktop Environment :: K Desktop Environment (KDE) :: Themes',
    'Topic :: Desktop Environment :: PicoGUI',
    'Topic :: Desktop Environment :: PicoGUI :: Applications',
    'Topic :: Desktop Environment :: PicoGUI :: Themes',
    'Topic :: Desktop Environment :: Screen Savers',
    'Topic :: Desktop Environment :: Window Managers',
    'Topic :: Desktop Environment :: Window Managers :: Afterstep',
    'Topic :: Desktop Environment :: Window Managers :: Afterstep :: Themes',
    'Topic :: Desktop Environment :: Window Managers :: Applets',
    'Topic :: Desktop Environment :: Window Managers :: Blackbox',
    'Topic :: Desktop Environment :: Window Managers :: Blackbox :: Themes',
    'Topic :: Desktop Environment :: Window Managers :: CTWM',
    'Topic :: Desktop Environment :: Window Managers :: CTWM :: Themes',
    'Topic :: Desktop Environment :: Window Managers :: Enlightenment',
    'Topic :: Desktop Environment :: Window Managers :: Enlightenment :: Epplets',
    'Topic :: Desktop Environment :: Window Managers :: Enlightenment :: Themes DR15',
    'Topic :: Desktop Environment :: Window Managers :: Enlightenment :: Themes DR16',
    'Topic :: Desktop Environment :: Window Managers :: Enlightenment :: Themes DR17',
    'Topic :: Desktop Environment :: Window Managers :: FVWM',
    'Topic :: Desktop Environment :: Window Managers :: FVWM :: Themes',
    'Topic :: Desktop Environment :: Window Managers :: Fluxbox',
    'Topic :: Desktop Environment :: Window Managers :: Fluxbox :: Themes',
    'Topic :: Desktop Environment :: Window Managers :: IceWM',
    'Topic :: Desktop Environment :: Window Managers :: IceWM :: Themes',
    'Topic :: Desktop Environment :: Window Managers :: MetaCity',
    'Topic :: Desktop Environment :: Window Managers :: MetaCity :: Themes',
    'Topic :: Desktop Environment :: Window Managers :: Oroborus',
    'Topic :: Desktop Environment :: Window Managers :: Oroborus :: Themes',
    'Topic :: Desktop Environment :: Window Managers :: Sawfish',
    'Topic :: Desktop Environment :: Window Managers :: Sawfish :: Themes 0.30',
    'Topic :: Desktop Environment :: Window Managers :: Sawfish :: Themes pre-0.30',
    'Topic :: Desktop Environment :: Window Managers :: Waimea',
    'Topic :: Desktop Environment :: Window Managers :: Waimea :: Themes',
    'Topic :: Desktop Environment :: Window Managers :: Window Maker',
    'Topic :: Desktop Environment :: Window Managers :: Window Maker :: Applets',
    'Topic :: Desktop Environment :: Window Managers :: Window Maker :: Themes',
    'Topic :: Desktop Environment :: Window Managers :: XFCE',
    'Topic :: Desktop Environment :: Window Managers :: XFCE :: Themes',
    'Topic :: Documentation',
    'Topic :: Documentation :: Sphinx',
    'Topic :: Education',
    'Topic :: Education :: Computer Aided Instruction (CAI)',
    'Topic :: Education :: Testing',
    'Topic :: Games/Entertainment',
    'Topic :: Games/Entertainment :: Arcade',
    'Topic :: Games/Entertainment :: Board Games',
    'Topic :: Games/Entertainment :: First Person Shooters',
    'Topic :: Games/Entertainment :: Fortune Cookies',
    'Topic :: Games/Entertainment :: Multi-User Dungeons (MUD)',
    'Topic :: Games/Entertainment :: Puzzle Games',
    'Topic :: Games/Entertainment :: Real Time Strategy',
    'Topic :: Games/Entertainment :: Role-Playing',
    'Topic :: Games/Entertainment :: Side-Scrolling/Arcade Games',
    'Topic :: Games/Entertainment :: Simulation',
    'Topic :: Games/Entertainment :: Turn Based Strategy',
    'Topic :: Home Automation',
    'Topic :: Internet',
    'Topic :: Internet :: File Transfer Protocol (FTP)',
    'Topic :: Internet :: Finger',
    'Topic :: Internet :: Log Analysis',
    'Topic :: Internet :: Name Service (DNS)',
    'Topic :: Internet :: Proxy Servers',
    'Topic :: Internet :: WAP',
    'Topic :: Internet :: WWW/HTTP',
    'Topic :: Internet :: WWW/HTTP :: Browsers',
    'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    'Topic :: Internet :: WWW/HTTP :: Dynamic Content :: CGI Tools/Libraries',
    'Topic :: Internet :: WWW/HTTP :: Dynamic Content :: Content Management System',
    'Topic :: Internet :: WWW/HTTP :: Dynamic Content :: Message Boards',
    'Topic :: Internet :: WWW/HTTP :: Dynamic Content :: News/Diary',
    'Topic :: Internet :: WWW/HTTP :: Dynamic Content :: Page Counters',
    'Topic :: Internet :: WWW/HTTP :: Dynamic Content :: Wiki',
    'Topic :: Internet :: WWW/HTTP :: HTTP Servers',
    'Topic :: Internet :: WWW/HTTP :: Indexing/Search',
    'Topic :: Internet :: WWW/HTTP :: Session',
    'Topic :: Internet :: WWW/HTTP :: Site Management',
    'Topic :: Internet :: WWW/HTTP :: Site Management :: Link Checking',
    'Topic :: Internet :: WWW/HTTP :: WSGI',
    'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
    'Topic :: Internet :: WWW/HTTP :: WSGI :: Middleware',
    'Topic :: Internet :: WWW/HTTP :: WSGI :: Server',
    'Topic :: Internet :: XMPP',
    'Topic :: Internet :: Z39.50',
    'Topic :: Multimedia',
    'Topic :: Multimedia :: Graphics',
    'Topic :: Multimedia :: Graphics :: 3D Modeling',
    'Topic :: Multimedia :: Graphics :: 3D Rendering',
    'Topic :: Multimedia :: Graphics :: Capture',
    'Topic :: Multimedia :: Graphics :: Capture :: Digital Camera',
    'Topic :: Multimedia :: Graphics :: Capture :: Scanners',
    'Topic :: Multimedia :: Graphics :: Capture :: Screen Capture',
    'Topic :: Multimedia :: Graphics :: Editors',
    'Topic :: Multimedia :: Graphics :: Editors :: Raster-Based',
    'Topic :: Multimedia :: Graphics :: Editors :: Vector-Based',
    'Topic :: Multimedia :: Graphics :: Graphics Conversion',
    'Topic :: Multimedia :: Graphics :: Presentation',
    'Topic :: Multimedia :: Graphics :: Viewers',
    'Topic :: Multimedia :: Sound/Audio',
    'Topic :: Multimedia :: Sound/Audio :: Analysis',
    'Topic :: Multimedia :: Sound/Audio :: CD Audio',
    'Topic :: Multimedia :: Sound/Audio :: CD Audio :: CD Playing',
    'Topic :: Multimedia :: Sound/Audio :: CD Audio :: CD Ripping',
    'Topic :: Multimedia :: Sound/Audio :: CD Audio :: CD Writing',
    'Topic :: Multimedia :: Sound/Audio :: Capture/Recording',
    'Topic :: Multimedia :: Sound/Audio :: Conversion',
    'Topic :: Multimedia :: Sound/Audio :: Editors',
    'Topic :: Multimedia :: Sound/Audio :: MIDI',
    'Topic :: Multimedia :: Sound/Audio :: Mixers',
    'Topic :: Multimedia :: Sound/Audio :: Players',
    'Topic :: Multimedia :: Sound/Audio :: Players :: MP3',
    'Topic :: Multimedia :: Sound/Audio :: Sound Synthesis',
    'Topic :: Multimedia :: Sound/Audio :: Speech',
    'Topic :: Multimedia :: Video',
    'Topic :: Multimedia :: Video :: Capture',
    'Topic :: Multimedia :: Video :: Conversion',
    'Topic :: Multimedia :: Video :: Display',
    'Topic :: Multimedia :: Video :: Non-Linear Editor',
    'Topic :: Office/Business',
    'Topic :: Office/Business :: Financial',
    'Topic :: Office/Business :: Financial :: Accounting',
    'Topic :: Office/Business :: Financial :: Investment',
    'Topic :: Office/Business :: Financial :: Point-Of-Sale',
    'Topic :: Office/Business :: Financial :: Spreadsheet',
    'Topic :: Office/Business :: Groupware',
    'Topic :: Office/Business :: News/Diary',
    'Topic :: Office/Business :: Office Suites',
    'Topic :: Office/Business :: Scheduling',
    'Topic :: Other/Nonlisted Topic',
    'Topic :: Printing',
    'Topic :: Religion',
    'Topic :: Scientific/Engineering',
    'Topic :: Scientific/Engineering :: Artificial Intelligence',
    'Topic :: Scientific/Engineering :: Artificial Life',
    'Topic :: Scientific/Engineering :: Astronomy',
    'Topic :: Scientific/Engineering :: Atmospheric Science',
    'Topic :: Scientific/Engineering :: Bio-Informatics',
    'Topic :: Scientific/Engineering :: Chemistry',
    'Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)',
    'Topic :: Scientific/Engineering :: GIS',
    'Topic :: Scientific/Engineering :: Human Machine Interfaces',
    'Topic :: Scientific/Engineering :: Hydrology',
    'Topic :: Scientific/Engineering :: Image Processing',
    'Topic :: Scientific/Engineering :: Image Recognition',
    'Topic :: Scientific/Engineering :: Information Analysis',
    'Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator',
    'Topic :: Scientific/Engineering :: Mathematics',
    'Topic :: Scientific/Engineering :: Medical Science Apps.',
    'Topic :: Scientific/Engineering :: Physics',
    'Topic :: Scientific/Engineering :: Visualization',
    'Topic :: Security',
    'Topic :: Security :: Cryptography',
    'Topic :: Sociology',
    'Topic :: Sociology :: Genealogy',
    'Topic :: Sociology :: History',
    'Topic :: Software Development',
    'Topic :: Software Development :: Assemblers',
    'Topic :: Software Development :: Bug Tracking',
    'Topic :: Software Development :: Build Tools',
    'Topic :: Software Development :: Code Generators',
    'Topic :: Software Development :: Compilers',
    'Topic :: Software Development :: Debuggers',
    'Topic :: Software Development :: Disassemblers',
    'Topic :: Software Development :: Documentation',
    'Topic :: Software Development :: Embedded Systems',
    'Topic :: Software Development :: Internationalization',
    'Topic :: Software Development :: Interpreters',
    'Topic :: Software Development :: Libraries',
    'Topic :: Software Development :: Libraries :: Application Frameworks',
    'Topic :: Software Development :: Libraries :: Java Libraries',
    'Topic :: Software Development :: Libraries :: PHP Classes',
    'Topic :: Software Development :: Libraries :: Perl Modules',
    'Topic :: Software Development :: Libraries :: Pike Modules',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Topic :: Software Development :: Libraries :: Ruby Modules',
    'Topic :: Software Development :: Libraries :: Tcl Extensions',
    'Topic :: Software Development :: Libraries :: pygame',
    'Topic :: Software Development :: Localization',
    'Topic :: Software Development :: Object Brokering',
    'Topic :: Software Development :: Object Brokering :: CORBA',
    'Topic :: Software Development :: Pre-processors',
    'Topic :: Software Development :: Quality Assurance',
    'Topic :: Software Development :: Testing',
    'Topic :: Software Development :: Testing :: Acceptance',
    'Topic :: Software Development :: Testing :: BDD',
    'Topic :: Software Development :: Testing :: Mocking',
    'Topic :: Software Development :: Testing :: Traffic Generation',
    'Topic :: Software Development :: Testing :: Unit',
    'Topic :: Software Development :: User Interfaces',
    'Topic :: Software Development :: Version Control',
    'Topic :: Software Development :: Version Control :: Bazaar',
    'Topic :: Software Development :: Version Control :: CVS',
    'Topic :: Software Development :: Version Control :: Git',
    'Topic :: Software Development :: Version Control :: Mercurial',
    'Topic :: Software Development :: Version Control :: RCS',
    'Topic :: Software Development :: Version Control :: SCCS',
    'Topic :: Software Development :: Widget Sets',
    'Topic :: System',
    'Topic :: System :: Archiving',
    'Topic :: System :: Archiving :: Backup',
    'Topic :: System :: Archiving :: Compression',
    'Topic :: System :: Archiving :: Mirroring',
    'Topic :: System :: Archiving :: Packaging',
    'Topic :: System :: Benchmark',
    'Topic :: System :: Boot',
    'Topic :: System :: Boot :: Init',
    'Topic :: System :: Clustering',
    'Topic :: System :: Console Fonts',
    'Topic :: System :: Distributed Computing',
    'Topic :: System :: Emulators',
    'Topic :: System :: Filesystems',
    'Topic :: System :: Hardware',
    'Topic :: System :: Hardware :: Hardware Drivers',
    'Topic :: System :: Hardware :: Mainframes',
    'Topic :: System :: Hardware :: Symmetric Multi-processing',
    'Topic :: System :: Hardware :: Universal Serial Bus (USB)',
    'Topic :: System :: Hardware :: Universal Serial Bus (USB) :: Audio',
    'Topic :: System :: Hardware :: Universal Serial Bus (USB) :: Audio/Video (AV)',
    'Topic :: System :: Hardware :: Universal Serial Bus (USB) :: Communications Device Class (CDC)',
    'Topic :: System :: Hardware :: Universal Serial Bus (USB) :: Diagnostic Device',
    'Topic :: System :: Hardware :: Universal Serial Bus (USB) :: Hub',
    'Topic :: System :: Hardware :: Universal Serial Bus (USB) :: Human Interface Device (HID)',
    'Topic :: System :: Hardware :: Universal Serial Bus (USB) :: Mass Storage',
    'Topic :: System :: Hardware :: Universal Serial Bus (USB) :: Miscellaneous',
    'Topic :: System :: Hardware :: Universal Serial Bus (USB) :: Printer',
    'Topic :: System :: Hardware :: Universal Serial Bus (USB) :: Smart Card',
    'Topic :: System :: Hardware :: Universal Serial Bus (USB) :: Vendor',
    'Topic :: System :: Hardware :: Universal Serial Bus (USB) :: Video (UVC)',
    'Topic :: System :: Hardware :: Universal Serial Bus (USB) :: Wireless Controller',
    'Topic :: System :: Installation/Setup',
    'Topic :: System :: Logging',
    'Topic :: System :: Monitoring',
    'Topic :: System :: Networking',
    'Topic :: System :: Networking :: Firewalls',
    'Topic :: System :: Networking :: Monitoring',
    'Topic :: System :: Networking :: Monitoring :: Hardware Watchdog',
    'Topic :: System :: Networking :: Time Synchronization',
    'Topic :: System :: Operating System',
    'Topic :: System :: Operating System Kernels',
    'Topic :: System :: Operating System Kernels :: BSD',
    'Topic :: System :: Operating System Kernels :: GNU Hurd',
    'Topic :: System :: Operating System Kernels :: Linux',
    'Topic :: System :: Power (UPS)',
    'Topic :: System :: Recovery Tools',
    'Topic :: System :: Shells',
    'Topic :: System :: Software Distribution',
    'Topic :: System :: System Shells',
    'Topic :: System :: Systems Administration',
    'Topic :: System :: Systems Administration :: Authentication/Directory',
    'Topic :: System :: Systems Administration :: Authentication/Directory :: LDAP',
    'Topic :: System :: Systems Administration :: Authentication/Directory :: NIS',
    'Topic :: Terminals',
    'Topic :: Terminals :: Serial',
    'Topic :: Terminals :: Telnet',
    'Topic :: Terminals :: Terminal Emulators/X Terminals',
    'Topic :: Text Editors',
    'Topic :: Text Editors :: Documentation',
    'Topic :: Text Editors :: Emacs',
    'Topic :: Text Editors :: Integrated Development Environments (IDE)',
    'Topic :: Text Editors :: Text Processing',
    'Topic :: Text Editors :: Word Processors',
    'Topic :: Text Processing',
    'Topic :: Text Processing :: Filters',
    'Topic :: Text Processing :: Fonts',
    'Topic :: Text Processing :: General',
    'Topic :: Text Processing :: Indexing',
    'Topic :: Text Processing :: Linguistic',
    'Topic :: Text Processing :: Markup',
    'Topic :: Text Processing :: Markup :: HTML',
    'Topic :: Text Processing :: Markup :: LaTeX',
    'Topic :: Text Processing :: Markup :: Markdown',
    'Topic :: Text Processing :: Markup :: SGML',
    'Topic :: Text Processing :: Markup :: VRML',
    'Topic :: Text Processing :: Markup :: XML',
    'Topic :: Text Processing :: Markup :: reStructuredText',
    'Topic :: Utilities',
    'Typing :: Stubs Only',
    'Typing :: Typed',
]
KNOWN_CLASSIFIERS = set(SORTED_CLASSIFIERS)


def is_private(classifier):
    return classifier.lower().startswith('private ::')



from hatchling.builders.custom import CustomBuilder
from hatchling.builders.sdist import SdistBuilder
from hatchling.builders.wheel import WheelBuilder
from hatchling.plugin import hookimpl


@hookimpl
def hatch_register_builder():
    return [CustomBuilder, SdistBuilder, WheelBuilder]



import hashlib
import os
import stat
import tempfile
import zipfile
from contextlib import closing
from io import StringIO

from hatchling.__about__ import __version__
from hatchling.builders.config import BuilderConfig
from hatchling.builders.plugin.interface import BuilderInterface
from hatchling.builders.utils import (
    format_file_hash,
    get_known_python_major_versions,
    get_reproducible_timestamp,
    normalize_archive_path,
    normalize_file_permissions,
    normalize_inclusion_map,
    replace_file,
    set_zip_info_mode,
)
from hatchling.metadata.spec import DEFAULT_METADATA_VERSION, get_core_metadata_constructors

try:
    from editables import EditableProject
except ImportError:  # no cov
    EditableProject = None

EDITABLES_MINIMUM_VERSION = '0.3'


class WheelArchive:
    def __init__(self, project_id, reproducible):
        """
        https://peps.python.org/pep-0427/#abstract
        """
        self.metadata_directory = f'{project_id}.dist-info'
        self.shared_data_directory = f'{project_id}.data'

        self.reproducible = reproducible
        if self.reproducible:
            self.time_tuple = self.get_reproducible_time_tuple()
        else:
            self.time_tuple = None

        raw_fd, self.path = tempfile.mkstemp(suffix='.whl')
        self.fd = os.fdopen(raw_fd, 'w+b')
        self.zf = zipfile.ZipFile(self.fd, 'w', compression=zipfile.ZIP_DEFLATED)

    @staticmethod
    def get_reproducible_time_tuple():
        from datetime import datetime, timezone

        d = datetime.fromtimestamp(get_reproducible_timestamp(), timezone.utc)
        return d.year, d.month, d.day, d.hour, d.minute, d.second

    def add_file(self, included_file):
        relative_path = normalize_archive_path(included_file.distribution_path)
        file_stat = os.stat(included_file.path)

        if self.reproducible:
            zip_info = zipfile.ZipInfo(relative_path, self.time_tuple)

            # https://github.com/takluyver/flit/pull/66
            new_mode = normalize_file_permissions(file_stat.st_mode)
            set_zip_info_mode(zip_info, new_mode & 0xFFFF)
            if stat.S_ISDIR(file_stat.st_mode):  # no cov
                zip_info.external_attr |= 0x10
        else:
            zip_info = zipfile.ZipInfo.from_file(included_file.path, relative_path)

        zip_info.compress_type = zipfile.ZIP_DEFLATED

        hash_obj = hashlib.sha256()
        with open(included_file.path, 'rb') as in_file, self.zf.open(zip_info, 'w') as out_file:
            while True:
                chunk = in_file.read(16384)
                if not chunk:
                    break

                hash_obj.update(chunk)
                out_file.write(chunk)

        hash_digest = format_file_hash(hash_obj.digest())
        return relative_path, hash_digest, file_stat.st_size

    def write_metadata(self, relative_path, contents):
        relative_path = f'{self.metadata_directory}/{normalize_archive_path(relative_path)}'
        return self.write_file(relative_path, contents)

    def add_shared_file(self, shared_file):
        shared_file.distribution_path = f'{self.shared_data_directory}/data/{shared_file.distribution_path}'
        return self.add_file(shared_file)

    def add_extra_metadata_file(self, extra_metadata_file):
        extra_metadata_file.distribution_path = (
            f'{self.metadata_directory}/extra_metadata/{extra_metadata_file.distribution_path}'
        )
        return self.add_file(extra_metadata_file)

    def write_file(self, relative_path, contents):
        if not isinstance(contents, bytes):
            contents = contents.encode('utf-8')

        time_tuple = self.time_tuple or (2020, 2, 2, 0, 0, 0)
        zip_info = zipfile.ZipInfo(relative_path, time_tuple)
        set_zip_info_mode(zip_info)

        hash_obj = hashlib.sha256(contents)
        hash_digest = format_file_hash(hash_obj.digest())
        self.zf.writestr(zip_info, contents, compress_type=zipfile.ZIP_DEFLATED)

        return relative_path, hash_digest, len(contents)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.zf.close()
        self.fd.close()


class WheelBuilderConfig(BuilderConfig):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.__include_defined = bool(
            self.target_config.get('include', self.build_config.get('include'))
            or self.target_config.get('packages', self.build_config.get('packages'))
            or self.target_config.get('only-include', self.build_config.get('only-include'))
        )
        self.__include = []
        self.__exclude = []
        self.__packages = []

        self.__core_metadata_constructor = None
        self.__shared_data = None
        self.__extra_metadata = None
        self.__strict_naming = None

    def set_default_file_selection(self):
        if self.__include or self.__exclude or self.__packages:
            return

        project_name = self.builder.normalize_file_name_component(self.builder.metadata.core.name)
        if os.path.isfile(os.path.join(self.root, project_name, '__init__.py')):
            self.__packages.append(project_name)
        elif os.path.isfile(os.path.join(self.root, 'src', project_name, '__init__.py')):
            self.__packages.append(f'src/{project_name}')
        elif os.path.isfile(os.path.join(self.root, f'{project_name}.py')):
            self.__packages.append(f'{project_name}.py')
        else:
            from glob import glob

            possible_namespace_packages = glob(os.path.join(self.root, '*', project_name, '__init__.py'))
            if len(possible_namespace_packages) == 1:
                relative_path = os.path.relpath(possible_namespace_packages[0], self.root)
                namespace = relative_path.split(os.sep)[0]
                self.__packages.append(namespace)
            else:
                self.__include.append('*.py')
                self.__exclude.append('test*')

    def default_include(self):
        if not self.__include_defined:
            self.set_default_file_selection()

        return self.__include

    def default_exclude(self):
        if not self.__include_defined:
            self.set_default_file_selection()

        return self.__exclude

    def default_packages(self):
        if not self.__include_defined:
            self.set_default_file_selection()

        return self.__packages

    @property
    def core_metadata_constructor(self):
        if self.__core_metadata_constructor is None:
            core_metadata_version = self.target_config.get('core-metadata-version', DEFAULT_METADATA_VERSION)
            if not isinstance(core_metadata_version, str):
                raise TypeError(
                    f'Field `tool.hatch.build.targets.{self.plugin_name}.core-metadata-version` must be a string'
                )

            constructors = get_core_metadata_constructors()
            if core_metadata_version not in constructors:
                raise ValueError(
                    f'Unknown metadata version `{core_metadata_version}` for field '
                    f'`tool.hatch.build.targets.{self.plugin_name}.core-metadata-version`. '
                    f'Available: {", ".join(sorted(constructors))}'
                )

            self.__core_metadata_constructor = constructors[core_metadata_version]

        return self.__core_metadata_constructor

    @property
    def shared_data(self):
        if self.__shared_data is None:
            shared_data = self.target_config.get('shared-data', {})
            if not isinstance(shared_data, dict):
                raise TypeError(f'Field `tool.hatch.build.targets.{self.plugin_name}.shared-data` must be a mapping')

            for i, (source, relative_path) in enumerate(shared_data.items(), 1):
                if not source:
                    raise ValueError(
                        f'Source #{i} in field `tool.hatch.build.targets.{self.plugin_name}.shared-data` '
                        f'cannot be an empty string'
                    )
                elif not isinstance(relative_path, str):
                    raise TypeError(
                        f'Path for source `{source}` in field '
                        f'`tool.hatch.build.targets.{self.plugin_name}.shared-data` must be a string'
                    )
                elif not relative_path:
                    raise ValueError(
                        f'Path for source `{source}` in field '
                        f'`tool.hatch.build.targets.{self.plugin_name}.shared-data` cannot be an empty string'
                    )

            self.__shared_data = normalize_inclusion_map(shared_data, self.root)

        return self.__shared_data

    @property
    def extra_metadata(self):
        if self.__extra_metadata is None:
            extra_metadata = self.target_config.get('extra-metadata', {})
            if not isinstance(extra_metadata, dict):
                raise TypeError(f'Field `tool.hatch.build.targets.{self.plugin_name}.extra-metadata` must be a mapping')

            for i, (source, relative_path) in enumerate(extra_metadata.items(), 1):
                if not source:
                    raise ValueError(
                        f'Source #{i} in field `tool.hatch.build.targets.{self.plugin_name}.extra-metadata` '
                        f'cannot be an empty string'
                    )
                elif not isinstance(relative_path, str):
                    raise TypeError(
                        f'Path for source `{source}` in field '
                        f'`tool.hatch.build.targets.{self.plugin_name}.extra-metadata` must be a string'
                    )
                elif not relative_path:
                    raise ValueError(
                        f'Path for source `{source}` in field '
                        f'`tool.hatch.build.targets.{self.plugin_name}.extra-metadata` cannot be an empty string'
                    )

            self.__extra_metadata = normalize_inclusion_map(extra_metadata, self.root)

        return self.__extra_metadata

    @property
    def strict_naming(self):
        if self.__strict_naming is None:
            if 'strict-naming' in self.target_config:
                strict_naming = self.target_config['strict-naming']
                if not isinstance(strict_naming, bool):
                    raise TypeError(
                        f'Field `tool.hatch.build.targets.{self.plugin_name}.strict-naming` must be a boolean'
                    )
            else:
                strict_naming = self.build_config.get('strict-naming', True)
                if not isinstance(strict_naming, bool):
                    raise TypeError('Field `tool.hatch.build.strict-naming` must be a boolean')

            self.__strict_naming = strict_naming

        return self.__strict_naming


class WheelBuilder(BuilderInterface):
    """
    Build a binary distribution (.whl file)
    """

    PLUGIN_NAME = 'wheel'

    def get_version_api(self):
        return {'standard': self.build_standard, 'editable': self.build_editable}

    def get_default_versions(self):
        return ['standard']

    def clean(self, directory, versions):
        for filename in os.listdir(directory):
            if filename.endswith('.whl'):
                os.remove(os.path.join(directory, filename))

    def build_standard(self, directory, **build_data):
        if 'tag' not in build_data:
            if build_data['infer_tag']:
                from packaging.tags import sys_tags

                best_matching_tag = next(sys_tags())
                tag_parts = (best_matching_tag.interpreter, best_matching_tag.abi, best_matching_tag.platform)
                build_data['tag'] = '-'.join(tag_parts)
            else:
                build_data['tag'] = self.get_default_tag()

        with WheelArchive(self.artifact_project_id, self.config.reproducible) as archive, closing(
            StringIO()
        ) as records:
            for included_file in self.recurse_included_files():
                record = archive.add_file(included_file)
                records.write(self.format_record(record))

            self.write_data(archive, records, build_data, build_data['dependencies'])

            records.write(f'{archive.metadata_directory}/RECORD,,\n')
            archive.write_metadata('RECORD', records.getvalue())

        target = os.path.join(directory, f"{self.artifact_project_id}-{build_data['tag']}.whl")

        replace_file(archive.path, target)
        return target

    def build_editable(self, directory, **build_data):
        if self.config.dev_mode_dirs:
            return self.build_editable_explicit(directory, **build_data)
        else:
            return self.build_editable_detection(directory, **build_data)

    def build_editable_detection(self, directory, **build_data):
        build_data['tag'] = self.get_default_tag()

        with WheelArchive(self.artifact_project_id, self.config.reproducible) as archive, closing(
            StringIO()
        ) as records:
            exposed_packages = {}
            for included_file in self.recurse_project_files():
                if not included_file.path.endswith('.py'):
                    continue

                relative_path = included_file.relative_path
                distribution_path = included_file.distribution_path
                path_parts = relative_path.split(os.sep)

                # Root file
                if len(path_parts) == 1:  # no cov
                    exposed_packages[os.path.splitext(relative_path)[0]] = os.path.join(self.root, relative_path)
                    continue

                # Root package
                root_module = path_parts[0]
                if distribution_path == relative_path:
                    exposed_packages[root_module] = os.path.join(self.root, root_module)
                else:
                    distribution_module = distribution_path.split(os.sep)[0]
                    try:
                        exposed_packages[distribution_module] = os.path.join(
                            self.root,
                            f'{relative_path[:relative_path.index(distribution_path)]}{distribution_module}',
                        )
                    except ValueError:
                        raise ValueError(
                            'Dev mode installations are unsupported when any path rewrite in the `sources` option '
                            'changes a prefix rather than removes it, see: '
                            'https://github.com/pfmoore/editables/issues/20'
                        ) from None

            editable_project = EditableProject(self.metadata.core.name, self.root)

            if self.config.dev_mode_exact:
                for module, relative_path in exposed_packages.items():
                    editable_project.map(module, relative_path)
            else:
                for relative_path in exposed_packages.values():
                    editable_project.add_to_path(os.path.dirname(relative_path))

            for filename, content in sorted(editable_project.files()):
                record = archive.write_file(filename, content)
                records.write(self.format_record(record))

            for included_file in self.recurse_forced_files(self.get_forced_inclusion_map(build_data)):
                record = archive.add_file(included_file)
                records.write(self.format_record(record))

            extra_dependencies = list(build_data['dependencies'])
            for dependency in editable_project.dependencies():
                if dependency == 'editables':
                    dependency += f'~={EDITABLES_MINIMUM_VERSION}'
                else:  # no cov
                    pass

                extra_dependencies.append(dependency)

            self.write_data(archive, records, build_data, extra_dependencies)

            records.write(f'{archive.metadata_directory}/RECORD,,\n')
            archive.write_metadata('RECORD', records.getvalue())

        target = os.path.join(directory, f"{self.artifact_project_id}-{build_data['tag']}.whl")

        replace_file(archive.path, target)
        return target

    def build_editable_explicit(self, directory, **build_data):
        build_data['tag'] = self.get_default_tag()

        with WheelArchive(self.artifact_project_id, self.config.reproducible) as archive, closing(
            StringIO()
        ) as records:
            directories = sorted(
                os.path.normpath(os.path.join(self.root, relative_directory))
                for relative_directory in self.config.dev_mode_dirs
            )

            record = archive.write_file(f"{self.metadata.core.name.replace('-', '_')}.pth", '\n'.join(directories))
            records.write(self.format_record(record))

            for included_file in self.recurse_forced_files(self.get_forced_inclusion_map(build_data)):
                record = archive.add_file(included_file)
                records.write(self.format_record(record))

            self.write_data(archive, records, build_data, build_data['dependencies'])

            records.write(f'{archive.metadata_directory}/RECORD,,\n')
            archive.write_metadata('RECORD', records.getvalue())

        target = os.path.join(directory, f"{self.artifact_project_id}-{build_data['tag']}.whl")

        replace_file(archive.path, target)
        return target

    def write_data(self, archive, records, build_data, extra_dependencies):
        self.add_shared_data(archive, records)

        # Ensure metadata is written last, see https://peps.python.org/pep-0427/#recommended-archiver-features
        self.write_metadata(archive, records, build_data, extra_dependencies=extra_dependencies)

    def add_shared_data(self, archive, records):
        for shared_file in self.recurse_explicit_files(self.config.shared_data):
            record = archive.add_shared_file(shared_file)
            records.write(self.format_record(record))

    def write_metadata(self, archive, records, build_data, extra_dependencies=()):
        # <<< IMPORTANT >>>
        # Ensure calls are ordered by the number of path components followed by the name of the components

        # METADATA
        self.write_project_metadata(archive, records, extra_dependencies=extra_dependencies)

        # WHEEL
        self.write_archive_metadata(archive, records, build_data)

        # entry_points.txt
        self.write_entry_points_file(archive, records)

        # licenses/
        self.add_licenses(archive, records)

        # extra_metadata/ - write last
        self.add_extra_metadata(archive, records)

    def write_archive_metadata(self, archive, records, build_data):
        from packaging.tags import parse_tag

        metadata = f"""\
Wheel-Version: 1.0
Generator: hatchling {__version__}
Root-Is-Purelib: {'true' if build_data['pure_python'] else 'false'}
"""

        for tag in sorted(map(str, parse_tag(build_data['tag']))):
            metadata += f'Tag: {tag}\n'

        record = archive.write_metadata('WHEEL', metadata)
        records.write(self.format_record(record))

    def write_entry_points_file(self, archive, records):
        entry_points_file = self.construct_entry_points_file()
        if entry_points_file:
            record = archive.write_metadata('entry_points.txt', entry_points_file)
            records.write(self.format_record(record))

    def write_project_metadata(self, archive, records, extra_dependencies=()):
        record = archive.write_metadata(
            'METADATA', self.config.core_metadata_constructor(self.metadata, extra_dependencies=extra_dependencies)
        )
        records.write(self.format_record(record))

    def add_licenses(self, archive, records):
        for relative_path in self.metadata.core.license_files:
            license_file = os.path.normpath(os.path.join(self.root, relative_path))
            with open(license_file, 'rb') as f:
                record = archive.write_metadata(f'licenses/{relative_path}', f.read())
                records.write(self.format_record(record))

    def add_extra_metadata(self, archive, records):
        for extra_metadata_file in self.recurse_explicit_files(self.config.extra_metadata):
            record = archive.add_extra_metadata_file(extra_metadata_file)
            records.write(self.format_record(record))

    def construct_entry_points_file(self):
        core_metadata = self.metadata.core
        metadata_file = ''

        if core_metadata.scripts:
            metadata_file += '\n[console_scripts]\n'
            for name, object_ref in core_metadata.scripts.items():
                metadata_file += f'{name} = {object_ref}\n'

        if core_metadata.gui_scripts:
            metadata_file += '\n[gui_scripts]\n'
            for name, object_ref in core_metadata.gui_scripts.items():
                metadata_file += f'{name} = {object_ref}\n'

        if core_metadata.entry_points:
            for group, entry_points in core_metadata.entry_points.items():
                metadata_file += f'\n[{group}]\n'
                for name, object_ref in entry_points.items():
                    metadata_file += f'{name} = {object_ref}\n'

        return metadata_file.lstrip()

    def get_default_tag(self):
        supported_python_versions = []
        for major_version in get_known_python_major_versions():
            for minor_version in range(100):
                if self.metadata.core.python_constraint.contains(f'{major_version}.{minor_version}'):
                    supported_python_versions.append(f'py{major_version}')
                    break

        return f'{".".join(supported_python_versions)}-none-any'

    def get_default_build_data(self):
        return {'infer_tag': False, 'pure_python': True, 'dependencies': [], 'force_include_editable': {}}

    def get_forced_inclusion_map(self, build_data):
        if not build_data['force_include_editable']:
            return self.config.get_force_include()

        return normalize_inclusion_map(build_data['force_include_editable'], self.root)

    @property
    def artifact_project_id(self):
        return (
            self.project_id
            if self.config.strict_naming
            else f'{self.normalize_file_name_component(self.metadata.core.raw_name)}-{self.metadata.version}'
        )

    @classmethod
    def get_config_class(cls):
        return WheelBuilderConfig

    @staticmethod
    def format_record(record):
        return '{},sha256={},{}\n'.format(*record)
